###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################


class _BaseException(Exception):
    """
    Inherits from 'Exception' and adds its own features

    Custom Base Exception that will generate an attribute called msg
    with the error message and will be used to catch errors.
    """
    ## Private attributes
    _args: tuple = None

    ## Public attributes
    msg: str = 'Application error.'

    ## Properties
    @property
    def args(self) -> tuple:
        """ Keeps the args attribute in sync with the msg attribute. """
        return self._args

    @args.setter
    def args(self, value: tuple) -> None:
        """  Used to keep sync the args and the msg attribute. """
        if not isinstance(value, tuple):
            raise ValueError(f"The 'args' value must be a tuple not {type(value)}.")

        self.msg = value[0]
        self._args = value

    ## Methods
    def __init__(self, *args: list, **kwargs: dict) -> None:
        super().__init__(*args)
        if len(args) == 1:
            self.msg = args[0]

        if kwargs:
            for attr, value in kwargs.items():
                setattr(self, attr, value)

    def __str__(self):
        return f'{self.msg}'

class DateError(_BaseException):
    """
    Custom exception class for date-related errors

    This exception class inherits from _BaseException and is used to handle errors related to date operations.


    Example:
        To raise a DateError exception:
        >>> raise DateError("Invalid date format.")
    """
    pass

class DefaultError(_BaseException):
    """
    Custom exception class for default errors.
    This exception class inherits from _BaseException and serves as a generic
    error class for handling default or unspecified errors.

    Usage:
        To raise a DefaultError exception:
        >>> raise DefaultError("An error occurred")
    """
    pass

class FieldValueError(_BaseException, ValueError):
    """
    Custom exception class for field value erros.
    This exception class inherits from both _BaseException and ValueError.
    It is used to handle errors related to invalid field values.

    Example:
        To raise a FieldValueError exception:
        >>> raise FieldValueError("Invalid field value.")
    """
    pass

class SDKValueError(_BaseException, ValueError):
    pass

class SDKTypeError(_BaseException, TypeError):
    pass

class EntityNotFound(_BaseException):
    pass

class HttpError(_BaseException):
    status_code: int = 500

    def __str__(self):
        return f'{self.status_code} -> {self.msg}'

class ReadonlyError(_BaseException):
    pass

class RedisEmptyListError(_BaseException):
    pass

class RequiredError(_BaseException):
    pass

class SDKError(_BaseException):
    pass

class QueryError(_BaseException):
    pass

class EntityError(_BaseException):
    pass

class InvalidArgumentError(_BaseException):
    pass

class APIError(_BaseException):
    # pylint: disable=import-outside-toplevel

    def __init__(self, code, message):
        from everysk.core.serialize import loads
        super().__init__(message)
        self.__code = code
        self.__message = loads(message, protocol='json') if message else message

    def __str__(self):
        from everysk.core.serialize import dumps
        if self.__code and self.__message:
            return dumps(self.__message, sort_keys=True, indent=2, protocol='json')
        return 'API ERROR'

class WorkerError(_BaseException):
    pass



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.core.datetime import Date, DateTime
from everysk.core.serialize import dumps, loads
from everysk.core.unittests import TestCase, mock


###############################################################################
#   Serialize Dumps Test Case Implementation
###############################################################################
class SerializeDumpsTestCase(TestCase):

    def test_int(self):
        self.assertEqual(dumps(1), '1')

    def test_float(self):
        self.assertEqual(dumps(1.1), '1.1')

    def test_bytes(self):
        self.assertEqual(dumps(b'bytes'), '"bytes"')

    def test_str(self):
        self.assertEqual(dumps('string'), '"string"')

    def test_list(self):
        self.assertEqual(dumps([1, 'string']), '[1, "string"]')

    def test_dict(self):
        self.assertEqual(dumps({'int': 1, 'str': 'string'}), '{"int": 1, "str": "string"}')

    def test_bool(self):
        self.assertEqual(dumps(True), 'true')

    def test_none(self):
        self.assertEqual(dumps(None), 'null')

    def test_date(self):
        self.assertEqual(dumps(Date(2023, 1, 1)), '{"__date__": "2023-01-01"}')

    def test_datetime(self):
        self.assertEqual(dumps(DateTime(2023, 1, 1, 12, 0, 0)), '{"__datetime__": "2023-01-01T12:00:00+00:00"}')

    def test_undefined(self):
        self.assertEqual(dumps(Undefined), 'null')

    def test_undefined_true(self):
        self.assertEqual(dumps(Undefined, use_undefined=True), '{"__undefined__": null}')

    def test_date_format(self):
        self.assertEqual(dumps(Date(2023, 1, 1), date_format='%Y%m%d'), '"20230101"')

    def test_datetime_format(self):
        self.assertEqual(
            dumps(DateTime(2023, 1, 1, 12, 0, 0), datetime_format='%Y%m%dT%H%M%S'),
            '"20230101T120000"'
        )

    def test_list_values(self):
        values = [1, 'string', Date(2023, 1, 1), DateTime(2023, 1, 1, 12, 0, 0), Undefined, None]
        self.assertEqual(
            dumps(values, use_undefined=True),
            '[1, "string", {"__date__": "2023-01-01"}, {"__datetime__": "2023-01-01T12:00:00+00:00"}, {"__undefined__": null}, null]'
        )

    def test_dict_values(self):
        values = {
            'int': 1,
            'str': 'string',
            'date': Date(2023, 1, 1),
            'datetime': DateTime(2023, 1, 1, 12, 0, 0),
            'undefined': Undefined,
            'none': None
        }
        self.assertEqual(
            dumps(values, use_undefined=True),
            '{"int": 1, "str": "string", "date": {"__date__": "2023-01-01"}, "datetime": {"__datetime__": "2023-01-01T12:00:00+00:00"}, "undefined": {"__undefined__": null}, "none": null}'
        )

    def test_protocol_json(self):
        self.assertEqual(dumps(1, protocol='json'), '1')

    def test_protocol_pickle(self):
        self.assertEqual(dumps(1, protocol='pickle'), b'\x80\x04K\x01.')

    def test_protocol_invalid(self):
        with self.assertRaisesRegex(ValueError, "Unsupported serialize protocol 'invalid'. Use 'json' or 'pickle'."):
            dumps(1, protocol='invalid')

    def test_allow_nan_true(self):
        self.assertEqual(dumps(float('nan'), allow_nan=True), 'NaN')
        self.assertEqual(dumps(float('inf'), allow_nan=True), 'Infinity')
        self.assertEqual(dumps(float('-inf'), allow_nan=True), '-Infinity')

    def test_allow_nan_false(self):
        with self.assertRaisesRegex(ValueError, 'Out of range float values are not JSON compliant'):
            dumps(float('nan'), allow_nan=False)
        with self.assertRaisesRegex(ValueError, 'Out of range float values are not JSON compliant'):
            dumps(float('inf'), allow_nan=False)
        with self.assertRaisesRegex(ValueError, 'Out of range float values are not JSON compliant'):
            dumps(float('-inf'), allow_nan=False)

    def test_object(self):
        class Test:
            def __init__(self, value):
                self.value = value

        with self.assertRaisesRegex(TypeError, 'Object of type Test is not JSON serializable'):
            dumps(Test(1))

    def test_sdk_objects(self):
        # pylint: disable=import-outside-toplevel
        from everysk.sdk.worker_base import WorkerBase
        worker_base = WorkerBase({})

        self.assertEqual(dumps(worker_base, sort_keys=True), '{"__class_path__": "everysk.sdk.worker_base.WorkerBase", "inputs_info": null, "parallel_info": null, "script_inputs": null, "worker_id": null, "worker_type": null, "workflow_execution_id": null, "workflow_id": null, "workspace": null}')

    def test_sdk_engines_objects(self):
        # pylint: disable=import-outside-toplevel
        from everysk.sdk.engines.cache import UserCache
        from everysk.sdk.engines.market_data import MarketData
        from everysk.sdk.engines.expression import Expression
        from everysk.sdk.engines.compliance import Compliance

        self.assertEqual(dumps(UserCache(), sort_keys=True), '{"__class_path__": "everysk.sdk.engines.cache.UserCache"}')
        self.assertEqual(dumps(MarketData(), sort_keys=True), '{"__class_path__": "everysk.sdk.engines.market_data.MarketData"}')
        self.assertEqual(dumps(Expression(), sort_keys=True), '{"__class_path__": "everysk.sdk.engines.expression.Expression"}')
        self.assertEqual(dumps(Compliance(), sort_keys=True), '{"__class_path__": "everysk.sdk.engines.compliance.Compliance"}')

    def test_sdk_entities_objects(self):
        # pylint: disable=import-outside-toplevel
        from everysk.sdk.entities.portfolio.base import Portfolio
        from everysk.sdk.entities.portfolio.security import Security
        from everysk.sdk.entities.portfolio.securities import Securities
        from everysk.sdk.entities.custom_index.base import CustomIndex
        from everysk.sdk.entities.datastore.base import Datastore
        from everysk.sdk.entities.file.base import File
        from everysk.sdk.entities.private_security.base import PrivateSecurity
        from everysk.sdk.entities.query import Query
        from everysk.sdk.entities.report.base import Report
        from everysk.sdk.entities.script import Script

        with mock.patch.object(DateTime, 'now', return_value=DateTime(2024, 6, 20, 14, 38, 59, 360554)):
            self.assertEqual(dumps(Portfolio(), sort_keys=True), '{"__class_path__": "everysk.sdk.entities.portfolio.base.Portfolio", "base_currency": null, "check_securities": false, "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "date": null, "description": null, "id": null, "level": null, "link_uid": null, "name": null, "nlv": null, "securities": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "version": "v1", "workspace": null}')
            self.assertEqual(dumps(Security(), sort_keys=True), '{"__class_path__": "everysk.sdk.entities.portfolio.security.Security", "accounting": null, "asset_class": null, "asset_subclass": null, "book": null, "comparable": null, "cost_price": null, "coupon": null, "currency": null, "display": null, "error_message": null, "error_type": null, "exchange": null, "extra_data": null, "fx_rate": null, "hash": null, "id": null, "indexer": null, "instrument_class": null, "instrument_subtype": null, "instrument_type": null, "isin": null, "issue_date": null, "issue_price": null, "issuer": null, "issuer_type": null, "label": null, "look_through_reference": null, "market_price": null, "market_value": null, "market_value_in_base": null, "maturity_date": null, "multiplier": null, "name": null, "operation": null, "option_type": null, "percent_index": null, "premium": null, "previous_quantity": null, "quantity": null, "rate": null, "return_date": null, "series": null, "settlement": null, "status": null, "strike": null, "symbol": null, "ticker": null, "trade_id": null, "trader": null, "underlying": null, "unrealized_pl": null, "unrealized_pl_in_base": null, "warranty": null}')
            self.assertEqual(dumps(Securities(), sort_keys=True), '[]')
            self.assertEqual(dumps(CustomIndex(), sort_keys=True), '{"__class_path__": "everysk.sdk.entities.custom_index.base.CustomIndex", "base_price": null, "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "currency": null, "data": null, "data_type": null, "description": null, "name": null, "periodicity": null, "symbol": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "version": "v1"}')
            self.assertEqual(dumps(Datastore(), sort_keys=True), '{"__class_path__": "everysk.sdk.entities.datastore.base.Datastore", "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "data": null, "date": null, "description": null, "id": null, "level": null, "link_uid": null, "name": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "version": "v1", "workspace": null}')
            self.assertEqual(dumps(File(), sort_keys=True), '{"__class_path__": "everysk.sdk.entities.file.base.File", "content_type": null, "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "data": null, "date": null, "description": null, "hash": null, "id": null, "link_uid": null, "name": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "url": null, "version": "v1", "workspace": null}')
            self.assertEqual(dumps(PrivateSecurity(), sort_keys=True), '{"__class_path__": "everysk.sdk.entities.private_security.base.PrivateSecurity", "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "currency": null, "data": null, "description": null, "instrument_type": null, "name": null, "symbol": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "version": "v1"}')
            self.assertEqual(dumps(Query(Portfolio), sort_keys=True), '{"__class_path__": "everysk.sdk.entities.query.Query", "_clean_order": [], "_klass": "Portfolio", "distinct_on": [], "filters": [], "limit": null, "offset": null, "order": [], "page_size": null, "page_token": null, "projection": null}')
            self.assertEqual(dumps(Report(), sort_keys=True), '{"__class_path__": "everysk.sdk.entities.report.base.Report", "authorization": null, "config_cascaded": null, "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "date": null, "description": null, "id": null, "layout_content": null, "link_uid": null, "name": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "url": null, "version": "v1", "widgets": null, "workspace": null}')
            self.assertEqual(dumps(Script(Portfolio), sort_keys=True), '{"__class_path__": "everysk.sdk.entities.script.Script", "_klass": "Portfolio"}')

    def test_base_object(self):
        # pylint: disable=import-outside-toplevel
        from everysk.core.object import BaseObject
        self.assertEqual(dumps(BaseObject(a=1, b={}, c=[3]), sort_keys=True), '{"__class_path__": "everysk.core.object.BaseObject", "a": 1, "b": {}, "c": [3]}')

    def test_base_dict(self):
        # pylint: disable=import-outside-toplevel
        from everysk.core.object import BaseDict
        self.assertEqual(dumps(BaseDict(a=1, b={}, c=[3]), sort_keys=True), '{"__class_path__": "everysk.core.object.BaseDict", "a": 1, "b": {}, "c": [3]}')

###############################################################################
#   Serialize Loads Test Case Implementation
###############################################################################
class SerializeLoadsTestCase(TestCase):

    def test_int(self):
        self.assertEqual(loads('1'), 1)

    def test_float(self):
        self.assertEqual(loads('1.1'), 1.1)

    def test_str(self):
        self.assertEqual(loads('"string"'), 'string')

    def test_list(self):
        self.assertEqual(loads('[1, "string"]'), [1, 'string'])

    def test_dict(self):
        self.assertEqual(loads('{"int": 1, "str": "string"}'), {'int': 1, 'str': 'string'})

    def test_bool(self):
        self.assertEqual(loads('true'), True)

    def test_none(self):
        self.assertIsNone(loads('null'))

    def test_date(self):
        self.assertEqual(loads('{"__date__": "2023-01-01"}'), Date(2023, 1, 1))

    def test_datetime(self):
        self.assertEqual(loads('{"__datetime__": "2023-01-01T12:00:00+00:00"}'), DateTime(2023, 1, 1, 12, 0, 0))

    def test_undefined_true(self):
        self.assertEqual(loads('{"__undefined__": null}', use_undefined=True), Undefined)

    def test_undefined_false(self):
        self.assertIsNone(loads('{"__undefined__": null}', use_undefined=False))

    def test_null(self):
        self.assertIsNone(loads('null'))

    def test_date_format(self):
        self.assertEqual(loads('{"__date__": "20230101"}', date_format='%Y%m%d'), Date(2023, 1, 1))

    def test_datetime_format(self):
        self.assertEqual(
            loads('{"__datetime__": "20230101T120000"}', datetime_format='%Y%m%dT%H%M%S'),
            DateTime(2023, 1, 1, 12, 0, 0)
        )

    def test_list_values(self):
        values = [1, 'string', Date(2023, 1, 1), DateTime(2023, 1, 1, 12, 0, 0), Undefined, None]
        self.assertEqual(
            loads(
                '[1, "string", {"__date__": "2023-01-01"}, {"__datetime__": "2023-01-01T12:00:00+00:00"}, {"__undefined__": null}, null]',
                use_undefined=True
            ),
            values
        )

    def test_dict_values(self):
        values = {
            'int': 1,
            'str': 'string',
            'date': Date(2023, 1, 1),
            'datetime': DateTime(2023, 1, 1, 12, 0, 0),
            'undefined': Undefined,
            'none': None
        }
        self.assertEqual(
            loads(
                '{"int": 1, "str": "string", "date": {"__date__": "2023-01-01"}, "datetime": {"__datetime__": "2023-01-01T12:00:00+00:00"}, "undefined": {"__undefined__": null}, "none": null}',
                use_undefined=True
            ),
            values
        )

    def test_protocol_json(self):
        self.assertEqual(loads('1', protocol='json'), 1)

    def test_protocol_pickle(self):
        self.assertEqual(loads(b'\x80\x04K\x01.', protocol='pickle'), 1)

    def test_protocol_invalid(self):
        with self.assertRaisesRegex(ValueError, "Unsupported serialize protocol 'invalid'. Use 'json' or 'pickle'."):
            loads('1', protocol='invalid')

    def test_nan(self):
        nan = loads('NaN')
        self.assertNotEqual(nan, nan)

    def test_inf(self):
        self.assertEqual(loads('Infinity'), float('inf'))
        self.assertEqual(loads('-Infinity'), float('-inf'))

    def test_bytes(self):
        self.assertEqual(loads(b'1'), 1)

    def test_sdk_objects(self):
        # pylint: disable=import-outside-toplevel
        from everysk.sdk.worker_base import WorkerBase

        self.assertEqual(
            loads('{"__class_path__": "app.var.sdk.worker_base.WorkerBase", "inputs_info": null, "parallel_info": null, "script_inputs": null, "worker_id": null, "worker_type": null, "workflow_execution_id": null, "workflow_id": null, "workspace": null}'),
            WorkerBase({})
        )

    def test_sdk_engines_objects(self):
        # pylint: disable=import-outside-toplevel
        from everysk.sdk.engines.cache import UserCache
        from everysk.sdk.engines.market_data import MarketData
        from everysk.sdk.engines.expression import Expression
        from everysk.sdk.engines.compliance import Compliance
        self.assertIsInstance(loads('{"__class_path__": "app.var.sdk.engines.cache.UserCache"}'), UserCache)
        self.assertIsInstance(loads('{"__class_path__": "app.var.sdk.engines.market_data.MarketData"}'), MarketData)
        self.assertIsInstance(loads('{"__class_path__": "app.var.sdk.engines.expression.Expression"}'), Expression)
        self.assertIsInstance(loads('{"__class_path__": "app.var.sdk.engines.compliance.Compliance"}'), Compliance)

    def test_sdk_entities_objects(self):
        # pylint: disable=import-outside-toplevel
        from everysk.sdk.entities.portfolio.base import Portfolio
        from everysk.sdk.entities.portfolio.security import Security
        from everysk.sdk.entities.portfolio.securities import Securities
        from everysk.sdk.entities.custom_index.base import CustomIndex
        from everysk.sdk.entities.datastore.base import Datastore
        from everysk.sdk.entities.file.base import File
        from everysk.sdk.entities.private_security.base import PrivateSecurity
        from everysk.sdk.entities.query import Query
        from everysk.sdk.entities.report.base import Report
        from everysk.sdk.entities.script import Script

        with mock.patch.object(DateTime, 'now', return_value=DateTime(2024, 6, 20, 14, 38, 59, 360554)):
            self.assertDictEqual(
                loads('{"__class_path__": "app.var.sdk.entities.portfolio.base.Portfolio", "base_currency": null, "check_securities": false, "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "date": null, "description": null, "id": null, "level": null, "link_uid": null, "name": null, "nlv": null, "securities": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "version": "v1", "workspace": null}'),
                Portfolio()
            )
            self.assertDictEqual(
                loads('{"__class_path__": "app.var.sdk.entities.portfolio.security.Security", "accounting": null, "asset_class": null, "asset_subclass": null, "book": null, "comparable": null, "cost_price": null, "coupon": null, "currency": null, "display": null, "error_message": null, "error_type": null, "exchange": null, "extra_data": null, "fx_rate": null, "hash": null, "id": null, "indexer": null, "instrument_class": null, "instrument_subtype": null, "instrument_type": null, "isin": null, "issue_date": null, "issue_price": null, "issuer": null, "issuer_type": null, "label": null, "look_through_reference": null, "market_price": null, "market_value": null, "market_value_in_base": null, "maturity_date": null, "multiplier": null, "name": null, "operation": null, "option_type": null, "percent_index": null, "premium": null, "previous_quantity": null, "quantity": null, "rate": null, "return_date": null, "series": null, "settlement": null, "status": null, "strike": null, "symbol": null, "ticker": null, "trade_id": null, "trader": null, "underlying": null, "unrealized_pl": null, "unrealized_pl_in_base": null, "warranty": null}'),
                Security()
            )
            self.assertEqual(loads('[]'), Securities())
            self.assertDictEqual(
                loads('{"__class_path__": "app.var.sdk.entities.custom_index.base.CustomIndex", "base_price": null, "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "currency": null, "data": null, "data_type": null, "description": null, "name": null, "periodicity": null, "symbol": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "version": "v1"}'),
                CustomIndex()
            )
            self.assertDictEqual(
                loads('{"__class_path__": "app.var.sdk.entities.datastore.base.Datastore", "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "data": null, "date": null, "description": null, "id": null, "level": null, "link_uid": null, "name": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "version": "v1", "workspace": null}'),
                Datastore()
            )
            self.assertDictEqual(
                loads('{"__class_path__": "app.var.sdk.entities.file.base.File", "content_type": null, "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "data": null, "date": null, "description": null, "hash": null, "id": null, "link_uid": null, "name": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "url": null, "version": "v1", "workspace": null}'),
                File()
            )
            self.assertDictEqual(
                loads('{"__class_path__": "app.var.sdk.entities.private_security.base.PrivateSecurity", "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "currency": null, "data": null, "description": null, "instrument_type": null, "name": null, "symbol": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "version": "v1"}'),
                PrivateSecurity()
            )
            self.assertDictEqual(
                loads('{"__class_path__": "app.var.sdk.entities.query.Query", "_clean_order": [], "_klass": "Portfolio", "distinct_on": [], "filters": [], "limit": null, "offset": null, "order": [], "page_size": null, "page_token": null, "projection": null}'),
                Query(Portfolio)
            )
            self.assertDictEqual(
                loads('{"__class_path__": "app.var.sdk.entities.report.base.Report", "authorization": null, "config_cascaded": null, "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "date": null, "description": null, "id": null, "layout_content": null, "link_uid": null, "name": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "url": null, "version": "v1", "widgets": null, "workspace": null}'),
                Report()
            )
            self.assertDictEqual(
                loads('{"__class_path__": "app.var.sdk.entities.script.Script", "_klass": "Portfolio"}'),
                Script(Portfolio)
            )

    def test_sdk_entities_when_instantiate_object_is_false(self):
        # pylint: disable=import-outside-toplevel
        from everysk.sdk.entities.portfolio.base import Portfolio
        with mock.patch.object(DateTime, 'now', return_value=DateTime(2024, 6, 20, 14, 38, 59, 360554)):
            p = Portfolio()
            result = dumps(p)
            res_loads = loads(result, instantiate_object=False)

            self.assertIsInstance(res_loads, dict)
            self.assertDictEqual(res_loads, Portfolio())

    def test_base_object(self):
        # pylint: disable=import-outside-toplevel
        from everysk.core.object import BaseObject
        ret = loads('{"__class_path__": "everysk.core.object.BaseObject", "a": 1, "b": {}, "c": [3]}')
        self.assertIsInstance(ret, BaseObject)
        self.assertEqual(ret.a, 1)
        self.assertEqual(ret.b, {})
        self.assertEqual(ret.c, [3])

        ret = loads('{"__class_path__": "app.var.test.Foo", "a": 1, "b": {}, "c": [3]}')
        self.assertIsInstance(ret, BaseObject)
        self.assertEqual(ret.a, 1)
        self.assertEqual(ret.b, {})
        self.assertEqual(ret.c, [3])

    def test_base_dict(self):
        # pylint: disable=import-outside-toplevel
        from everysk.core.object import BaseDict
        self.assertDictEqual(loads('{"__class_path__": "everysk.core.object.BaseDict", "a": 1, "b": {}, "c": [3]}'), BaseDict(a=1, b={}, c=[3]))



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
import string
import secrets
import uuid

from everysk.config import settings

###############################################################################
#   Globals
###############################################################################
ALPHANUMERIC = string.ascii_letters + string.digits
SYMBOL_ID_MAX_LEN = settings.SYMBOL_ID_MAX_LEN

###############################################################################
#   Cryptography Implementation
###############################################################################
def generate_random_id(length: int, characters: str = ALPHANUMERIC) -> str:
    """
    Generate a random ID of the specified length.

    Args:
        length (int): The desired length for the random ID.
        characters (str, optional): The characters to use for generating the ID.
            Default is alphanumeric characters.

    Returns:
        str: The generated random ID consisting of the specified characters.

    Usage:
        >>> from everysk.core.engines.cryptography import generate_random_id
        >>> generate_random_id(18)
        >>> 'xiXwQuig6cI11ny095'
    """
    return ''.join(secrets.choice(characters) for i in range(length))

def generate_unique_id():
    """
    Generate a unique ID with fixed length (32 characters).

    Returns:
        str: The generated unique ID consisting of hexadecimal characters.

    Usage:
        >>> from everysk.sdk.engines.cryptography import generate_unique_id
        >>> generate_unique_id()
        >>> 'dbdc803042ff48e19a344ad080a102dd'

    Notes:
        UUIDs generated by this function are highly unlikely to collide
    """
    return uuid.uuid4().hex

def generate_short_random_id() -> str:
    """
    Generate a short random ID with fixed length (8 characters).

    Returns:
        str: The generated unique ID consisting of alphanumeric characters.

    Usage:
        >>> from everysk.sdk.engines.cryptography import generate_short_random_id
        >>> generate_short_random_id()
        >>> 'u0tmNqHP'

    Notes:
        The length of the generated ID is fixed to 8 characters
    """
    return generate_random_id(length=settings.SIMPLE_UNIQUE_ID_LENGTH)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
import json
import zoneinfo

from everysk.config import settings
from everysk.core.compress import compress
from everysk.core.datetime import DateTime, timezone
from everysk.core.exceptions import RequiredError, SDKValueError
from everysk.core.fields import StrField
from everysk.core.http import HttpSDKPOSTConnection
from everysk.core.unittests import TestCase, mock

from everysk.sdk.entities.query import Query
from everysk.sdk.entities.base import BaseEntity
from everysk.sdk.entities.portfolio.base import Portfolio
from everysk.sdk.entities.portfolio.securities import Securities


###############################################################################
#   Base Entity TestCase Implementation
###############################################################################
class TestBaseEntity(TestCase):

    def setUp(self):
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()
        self.expected_response = {
            '_klass': Portfolio,
            'filters': [],
            'order': [],
            'projection': None,
            'limit': None,
            'offset': None,
            'page_size': None,
            'page_token': None,
            'distinct_on': []
        }
        self.sample_portfolio_data = {
            'id': 'port_1234567891011211234567890',
            'workspace': 'SampleWorkspace',
            'name': 'SamplePortfolio',
            'tags': ['tag1', 'tag2'],
            'link_uid': 'ABC',
            'description': 'Description',
            'nlv': 1000.0,
            'base_currency': 'USD',
            'date': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'securities': Securities([{'symbol': 'AAPL'}]),
            'version': '1.0',
            'created_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'updated_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'level': 'HIGH',
            'outstanding_shares': 1000,
            'source_hash': 'XYZ',
            'status': 'OK',
            'portfolio_uid': 'UID123',
            'check_securities': False
        }
        self.portfolio = Portfolio(**self.sample_portfolio_data)

    def test_entity_to_query(self):
        portfolio = self.portfolio.copy()
        self.assertRaisesRegex(
            SDKValueError,
            "Can't filter by Name and Tags at the same time",
            portfolio.to_query
        )

        # test with name
        portfolio.update(tags=None, link_uid=None, date=None)
        _query = portfolio.to_query() # pylint: disable=protected-access
        self.assertDictEqual(_query,
        {
            '_klass': Portfolio,
            'filters': [
                ('workspace', '=', 'SampleWorkspace'),
                ('name', '<', 'sampleportfolip'),
                ('name', '>=', 'sampleportfolio')
            ],
            'order': [],
            'projection': None,
            'limit': None,
            'offset': None,
            'page_size': None,
            'page_token': None,
            'distinct_on': []
        })

        # test with tags
        portfolio.update(tags=['tag1', 'tag2'], name=None)
        _query = portfolio.to_query() # pylint: disable=protected-access
        self.assertListEqual(_query.filters, [
            ('workspace', '=', 'SampleWorkspace'),
            ('tags', '=', 'tag1'),
            ('tags', '=', 'tag2')
        ])

        # test with link_uid
        portfolio.update(tags=None, link_uid='ABC')
        _query = portfolio.to_query() # pylint: disable=protected-access
        self.assertListEqual(_query.filters, [
            ('workspace', '=', 'SampleWorkspace'),
            ('link_uid', '=', 'ABC')
        ])

    def test_check_query_raises_sdk_error_when_filtering_by_name_and_tags(self):
        with self.assertRaises(SDKValueError) as context:
            portfolio : Portfolio = self.portfolio.copy()
            portfolio.to_query(limit=1)

        self.assertEqual(str(context.exception), "Can't filter by Name and Tags at the same time")

    def test_check_query_raises_sdk_error_when_filtering_by_name_and_link(self):
        with self.assertRaises(SDKValueError) as context:
            portfolio: Portfolio = self.portfolio.copy()
            portfolio['tags'] = []
            portfolio.to_query(limit=1)

        self.assertEqual(str(context.exception), "Can't filter by Name and Link UID at the same time")

    def test_check_query_raises_sdk_error_when_date_in_query_order(self):
        with self.assertRaises(SDKValueError) as context:
            portfolio: Portfolio = self.portfolio.copy()
            portfolio['tags'] = []
            portfolio['link_uid'] = ''
            portfolio.to_query(order=['date'])

        self.assertEqual(str(context.exception), "Can't filter by Name and Date at the same time,  must order by updated_on")

    def test_get_id_prefix(self):
        with self.assertRaises(NotImplementedError):
            BaseEntity.get_id_prefix()

    def test_generate_id(self):
        with mock.patch.object(BaseEntity, 'get_id_prefix', return_value='P'), \
                mock.patch('everysk.sdk.entities.base.generate_random_id', return_value='12345'):
            self.assertEqual(BaseEntity.generate_id(), 'P12345')

    def test_validate_id(self):
        class FakeEntityClass(BaseEntity): # pylint: disable=abstract-method
            id = StrField(regex=settings.PORTFOLIO_ID_REGEX, max_size=settings.PORTFOLIO_ID_MAX_SIZE, required_lazy=True, empty_is_none=True)

        self.assertFalse(FakeEntityClass.validate_id('invalid_id'))
        self.assertTrue(FakeEntityClass.validate_id('port_puvGxrtIYc7OupBzfnrwFNVPg'))
        self.assertFalse(FakeEntityClass.validate_id(''))
        self.assertFalse(BaseEntity.validate_id(None))

    def test_validate(self):
        self.assertTrue(BaseEntity(id="my_id", workspace="my_workspace", name="my_name", created_on=DateTime.now(), updated_on=DateTime.now()).validate())

    def test_validate_transient(self):
        class FakeEntity(BaseEntity):
            @staticmethod
            def get_id_prefix():
                return 'fake_'

        entity = FakeEntity(workspace='my_workspace', name='my_name')
        entity['created_on'] = DateTime(2023, 9, 9, 9, 9, 9, 9)
        entity['updated_on'] = DateTime(2023, 9, 9, 9, 9, 9, 9)

        self.assertTrue(FakeEntity.validate_transient(entity))
        self.assertDictEqual(
            entity.__dict__,
            {'version': 'v1', 'created_on': DateTime(2023, 9, 9, 9, 9, 9, 9, tzinfo=zoneinfo.ZoneInfo(key='UTC')), 'updated_on': DateTime(2023, 9, 9, 9, 9, 9, 9, tzinfo=zoneinfo.ZoneInfo(key='UTC')), 'workspace': 'my_workspace', 'name': 'my_name'}
        )

    def test_raises_validate(self):
        base_entity = BaseEntity()
        base_entity['created_on'] = None
        self.assertRaisesRegex(
            RequiredError,
            'The created_on attribute is required.',
            base_entity.validate
        )

    def test_get_query(self):
        """ Test _get_query method """
        expected_response = self.expected_response.copy()
        expected_response['projection'] = ['securities']
        result = Portfolio.query.set_projection('securities') # pylint: disable=protected-access
        self.assertDictEqual(result, expected_response)

    def test_query(self):
        """ Test query method """
        expected_response = self.expected_response.copy()
        expected_response['projection'] = ['securities']
        result = Portfolio.query.set_projection('securities')
        self.assertDictEqual(result, expected_response)

    def test_where(self):
        """ Test where method """
        expected_response = self.expected_response.copy()
        expected_response['filters'] = [('date', '=', DateTime(2023, 8, 10, 12, 0, tzinfo=timezone.utc))]
        result = Portfolio.query.where('date', '=', '2023-08-10')
        self.assertDictEqual(result, expected_response)

    def test_sort_by(self):
        """ Test sort_by method """
        expected_response = self.expected_response.copy()
        expected_response['order'] = ['date']
        result = Portfolio.query.sort_by('date')
        self.assertDictEqual(result, expected_response)

    def test_set_projection(self):
        """ Test set_projection method """
        expected_response = self.expected_response.copy()
        expected_response['projection'] = ['securities']
        result = Portfolio.query.set_projection('securities')
        self.assertDictEqual(result, expected_response)

    def test_set_limit(self):
        """ Test set_limit method """
        expected_response = self.expected_response.copy()
        expected_response['limit'] = 5
        result = Portfolio.query.set_limit(5)
        self.assertDictEqual(result, expected_response)

    def test_set_offset(self):
        """ Test set_offset method """
        expected_response = self.expected_response.copy()
        expected_response['offset'] = 10
        result = Portfolio.query.set_offset(10)
        self.assertDictEqual(result, expected_response)

    def test_set_page_size(self):
        """ Test set_page_size method """
        expected_response = self.expected_response.copy()
        expected_response['page_size'] = 15
        result = Portfolio.query.set_page_size(15)
        self.assertDictEqual(result, expected_response)

    def test_set_page_token(self):
        """ Test set_page_token method """
        expected_response = self.expected_response.copy()
        expected_response['page_token'] = 'my_test_page'
        result = Portfolio.query.set_page_token('my_test_page')
        self.assertDictEqual(result, expected_response)

    def test_load(self):
        expected_data = compress({'class_name': 'Query', 'method_name': 'load', 'self_obj': {'_klass': 'Portfolio', 'filters': [['workspace', '=', 'my_workspace']], 'order': [], 'projection': None, 'distinct_on': [], 'limit': None, 'offset': None, 'page_size': None, 'page_token': None, '_clean_order': [], '__class_path__': 'everysk.sdk.entities.query.Query'}, 'params': {'offset': 5}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{"key": "value"}'
            mock_post.return_value.status_code = 200
            result = Portfolio.query.where('workspace', 'my_workspace').load(offset=5)

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)
        self.assertIsInstance(result, Portfolio)

    def test_list(self):
        expected_data = compress({'class_name': 'Query', 'method_name': 'loads', 'self_obj': {'_klass': 'Portfolio', 'filters': [['workspace', '=', 'my_workspace']], 'order': [], 'projection': None, 'distinct_on': [], 'limit': None, 'offset': None, 'page_size': None, 'page_token': None, '_clean_order': [], '__class_path__': 'everysk.sdk.entities.query.Query'}, 'params': {'limit': 10, 'offset': 5}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '[{"key": "value"}]'
            mock_post.return_value.status_code = 200
            result = Portfolio.query.where('workspace', 'my_workspace').loads(limit=10, offset=5)

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 1)
        self.assertIsInstance(result[0], Portfolio)

    def test_page(self):
        query_class = Query(**self.expected_response)
        query_class['filters'] = [('workspace', '=', 'my_workspace')]

        expected_data = compress({'class_name': 'Query', 'method_name': 'page', 'self_obj': {'_klass': 'Portfolio', 'filters': [['workspace', '=', 'my_workspace']], 'order': [], 'projection': None, 'distinct_on': [], 'limit': None, 'offset': None, 'page_size': None, 'page_token': None, '_clean_order': [], '__class_path__': 'everysk.sdk.entities.query.Query'}, 'params': {'page_size': 10, 'page_token': 'page_token'}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = json.dumps({
                "entities": [
                    {"key1": "value1", "key2": "value2", "key3": "value3"},
                ]
            })
            mock_post.return_value.status_code = 200
            result = Portfolio.query.where('workspace', 'my_workspace').page(page_size=10, page_token='page_token')

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)
        self.assertIsInstance(result['entities'][0], Portfolio)

    def test_pages(self):
        query = Query(**self.expected_response)
        query['filters'] = [('workspace', '=', 'my_workspace')]
        expected_data = compress({'class_name': 'Query', 'method_name': 'page', 'self_obj': {'_klass': 'Portfolio', 'filters': [['workspace', '=', 'my_workspace']], 'order': [], 'projection': None, 'distinct_on': [], 'limit': None, 'offset': None, 'page_size': None, 'page_token': None, '_clean_order': [], '__class_path__': 'everysk.sdk.entities.query.Query'}, 'params': {'page_size': 10, 'page_token': {'__undefined__': None}}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = json.dumps({
                "entities": [
                    {"key1": "value1", "key2": "value2", "key3": "value3"},
                    {"key1": "value4", "key2": "value5", "key3": "value6"},
                    {"key1": "value7", "key2": "value8", "key3": "value9"}
                ]
            })
            mock_post.return_value.status_code = 200
            result = Portfolio.query.where('workspace', 'my_workspace').pages(page_size=10)
            next(result)

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)

    def test_script_query(self):
        expected_data = compress({'class_name': 'Script', 'method_name': 'fetch', 'self_obj': {'_klass': 'Portfolio', '__class_path__': 'everysk.sdk.entities.script.Script'}, 'params': {'user_input': 'port_12456', 'variant': 'select', 'workspace': None}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = Portfolio.script.fetch(user_input='port_12456', variant='select', workspace=None)

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)
        self.assertIsInstance(result, Portfolio)

    def test_retrieve(self):
        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'retrieve', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890'}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{"key": "value"}'
            mock_post.return_value.status_code = 200
            Portfolio.retrieve(entity_id=self.sample_portfolio_data['id'])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)

    def test_retrieve_with_none(self):
        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'retrieve', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890'}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = 'null'
            mock_post.return_value.status_code = 200
            Portfolio.retrieve(entity_id=self.sample_portfolio_data['id'])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)

    def test_create(self):
        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'create', 'self_obj': None, 'params': {'entity_dict': {'id': 'port_1234567891011211234567890', 'workspace': 'SampleWorkspace', 'name': 'SamplePortfolio', 'tags': ['tag1', 'tag2'], 'link_uid': 'ABC', 'description': 'Description', 'nlv': 1000.0, 'base_currency': 'USD', 'date': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'securities': [], 'version': '1.0', 'created_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'updated_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'level': 'HIGH', 'outstanding_shares': 1000, 'source_hash': 'XYZ', 'status': 'OK', 'portfolio_uid': 'UID123', 'check_securities': False}}}, protocol='gzip', serialize='json')
        self.sample_portfolio_data['securities'] = []

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{"creating": "portfolio"}'
            mock_post.return_value.status_code = 200
            Portfolio.create(entity_dict=self.sample_portfolio_data)

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)

    def test_modify(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['base_currency'] = 'BRL'
        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'modify', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890', 'overwrites': {'base_currency': 'BRL'}}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            Portfolio.modify(entity_id=self.sample_portfolio_data['id'], overwrites={'base_currency': 'BRL'})

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)

    def test_modify_return_none(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['base_currency'] = 'BRL'

        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'modify', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890', 'overwrites': {'base_currency': 'BRL'}}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = 'null'
            mock_post.return_value.status_code = 200
            Portfolio.modify(entity_id=self.sample_portfolio_data['id'], overwrites={'base_currency': 'BRL'})

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)

    def test_remove(self):
        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'remove', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890'}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{"remove": "key"}'
            mock_post.return_value.status_code = 200
            result = Portfolio.remove(entity_id=self.sample_portfolio_data['id'])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)
        self.assertIsInstance(result, Portfolio)

    def test_remove_return_none(self):
        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'remove', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890'}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = 'null'
            mock_post.return_value.status_code = 200
            result = Portfolio.remove(entity_id=self.sample_portfolio_data['id'])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)
        self.assertIsNone(result)

    def test_clone(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['base_currency'] = 'BRL'
        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'clone', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890', 'overwrites': {'base_currency': 'BRL'}}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{"key": "value"}'
            mock_post.return_value.status_code = 200
            result = Portfolio.clone(entity_id=self.sample_portfolio_data['id'], overwrites={'base_currency': 'BRL'})

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)
        self.assertIsInstance(result, Portfolio)

    def test_clone_return_none(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['base_currency'] = 'BRL'
        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'clone', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890', 'overwrites': {'base_currency': 'BRL'}}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = 'null'
            mock_post.return_value.status_code = 200
            result = Portfolio.clone(entity_id=self.sample_portfolio_data['id'], overwrites={'base_currency': 'BRL'})

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)
        self.assertIsNone(result)

    def test_create_many(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['base_currency'] = 'BRL'
        sample_portfolio_data['check_securities'] = True
        sample_portfolio_data['securities'] = []
        self.sample_portfolio_data['securities'] = []

        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'create_many', 'self_obj': None, 'params': {'entity_dict_list': [{'id': 'port_1234567891011211234567890', 'workspace': 'SampleWorkspace', 'name': 'SamplePortfolio', 'tags': ['tag1', 'tag2'], 'link_uid': 'ABC', 'description': 'Description', 'nlv': 1000.0, 'base_currency': 'USD', 'date': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'securities': [], 'version': '1.0', 'created_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'updated_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'level': 'HIGH', 'outstanding_shares': 1000, 'source_hash': 'XYZ', 'status': 'OK', 'portfolio_uid': 'UID123', 'check_securities': False}, {'id': 'port_1234567891011211234567890', 'workspace': 'SampleWorkspace', 'name': 'SamplePortfolio', 'tags': ['tag1', 'tag2'], 'link_uid': 'ABC', 'description': 'Description', 'nlv': 1000.0, 'base_currency': 'BRL', 'date': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'securities': [], 'version': '1.0', 'created_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'updated_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'level': 'HIGH', 'outstanding_shares': 1000, 'source_hash': 'XYZ', 'status': 'OK', 'portfolio_uid': 'UID123', 'check_securities': True}]}}, protocol='gzip', serialize='json')
        expected_response = []

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = Portfolio.create_many(entity_dict_list=[self.sample_portfolio_data, sample_portfolio_data])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)
        self.assertEqual(result, expected_response)

    def test_modify_many(self):
        original_portfolio_data = self.sample_portfolio_data.copy()
        original_portfolio_data['name'] = 'modified_1'
        Portfolio(**original_portfolio_data)

        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['id'] = 'port_1234567891011211234567891'
        sample_portfolio_data['name'] = 'modified_2'
        Portfolio(**sample_portfolio_data)

        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'modify_many', 'self_obj': None, 'params': {'entity_id_list': ['port_1234567891011211234567890', 'port_1234567891011211234567891'], 'overwrites': [{'name': 'modified_1'}, {'name': 'modified_2'}]}}, protocol='gzip', serialize='json')
        expected_response = []

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = Portfolio.modify_many(entity_id_list=[self.sample_portfolio_data['id'], sample_portfolio_data['id']], overwrites=[{'name': 'modified_1'}, {'name': 'modified_2'}])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)
        self.assertEqual(result, expected_response)

    def test_remove_many(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['id'] = 'port_1234567891011211234567891'
        sample_portfolio_data['base_currency'] = 'BRL'

        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'remove_many', 'self_obj': None, 'params': {'entity_id_list': ['port_1234567891011211234567890', 'port_1234567891011211234567891']}}, protocol='gzip', serialize='json')
        expected_response = [self.sample_portfolio_data['id'], sample_portfolio_data['id']]

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = f'["{self.sample_portfolio_data["id"]}", "{sample_portfolio_data["id"]}"]'
            mock_post.return_value.status_code = 200
            result = Portfolio.remove_many(entity_id_list=[self.sample_portfolio_data['id'], sample_portfolio_data['id']])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)
        self.assertEqual(result, expected_response)

    def test_clone_many(self):
        original_portfolio_data = self.sample_portfolio_data.copy()
        original_portfolio_data['name'] = 'modified_1'
        original_portfolio_data['workspace'] = 'modified_1'
        original_portfolio = Portfolio(**original_portfolio_data)

        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['id'] = 'port_1234567891011211234567891'
        sample_portfolio_data['name'] = 'modified_2'
        sample_portfolio_data['workspace'] = 'modified_2'
        Portfolio(**sample_portfolio_data)

        entity_id_list = [original_portfolio['id'], sample_portfolio_data['id']]

        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'clone_many', 'self_obj': None, 'params': {'entity_id_list': ['port_1234567891011211234567890', 'port_1234567891011211234567891'], 'overwrites': [{'name': 'modified_1'}, {'name': 'modified_1'}]}}, protocol='gzip', serialize='json')
        expected_response = []

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = Portfolio.clone_many(entity_id_list=entity_id_list, overwrites=[{'name': 'modified_1'}, {'name': 'modified_1'}])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)
        self.assertEqual(result, expected_response)

    def test_save(self):
        self.sample_portfolio_data['securities'] = []
        portfolio = Portfolio(**self.sample_portfolio_data)
        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'save', 'self_obj': {'id': 'port_1234567891011211234567890', 'workspace': 'SampleWorkspace', 'name': 'SamplePortfolio', 'tags': ['tag1', 'tag2'], 'link_uid': 'ABC', 'description': 'Description', 'nlv': 1000.0, 'base_currency': 'USD', 'date': '20230909', 'securities': [], 'version': '1.0', 'created_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'updated_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'level': 'HIGH', 'outstanding_shares': 1000, 'source_hash': 'XYZ', 'status': 'OK', 'portfolio_uid': 'UID123', 'check_securities': False, '__class_path__': 'everysk.sdk.entities.portfolio.base.Portfolio'}, 'params': {}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{"key": "value"}'
            mock_post.return_value.status_code = 200
            result = portfolio.save()

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)
        self.assertIsInstance(result, Portfolio)

    def test_delete(self):
        self.sample_portfolio_data['securities'] = []
        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'delete', 'self_obj': {'id': 'port_1234567891011211234567890', 'workspace': 'SampleWorkspace', 'name': 'SamplePortfolio', 'tags': ['tag1', 'tag2'], 'link_uid': 'ABC', 'description': 'Description', 'nlv': 1000.0, 'base_currency': 'USD', 'date': '20230909', 'securities': [], 'version': '1.0', 'created_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'updated_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'level': 'HIGH', 'outstanding_shares': 1000, 'source_hash': 'XYZ', 'status': 'OK', 'portfolio_uid': 'UID123', 'check_securities': False, '__class_path__': 'everysk.sdk.entities.portfolio.base.Portfolio'}, 'params': {}}, protocol='gzip', serialize='json')
        portfolio = Portfolio(**self.sample_portfolio_data)

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = portfolio.delete()

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)
        self.assertIsInstance(result, Portfolio)

    def test_delete_return_none(self):
        self.sample_portfolio_data['securities'] = []
        portfolio = Portfolio(**self.sample_portfolio_data)
        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'delete', 'self_obj': {'id': 'port_1234567891011211234567890', 'workspace': 'SampleWorkspace', 'name': 'SamplePortfolio', 'tags': ['tag1', 'tag2'], 'link_uid': 'ABC', 'description': 'Description', 'nlv': 1000.0, 'base_currency': 'USD', 'date': '20230909', 'securities': [], 'version': '1.0', 'created_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'updated_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'level': 'HIGH', 'outstanding_shares': 1000, 'source_hash': 'XYZ', 'status': 'OK', 'portfolio_uid': 'UID123', 'check_securities': False, '__class_path__': 'everysk.sdk.entities.portfolio.base.Portfolio'}, 'params': {}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = 'null'
            mock_post.return_value.status_code = 200
            result = portfolio.delete()

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, verify=settings.HTTP_DEFAULT_SSL_VERIFY, timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT, data=expected_data)
        self.assertIsNone(result)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=protected-access
import sys
from logging import Logger as PythonLogger, DEBUG, INFO, StreamHandler
from everysk.config import settings
from everysk.core import log as log_module, slack as slack_module
from everysk.core.log import Logger, LoggerManager
from everysk.core.unittests import TestCase, mock


class LoggerManagerTestCase(TestCase):

    def test_context_manager(self):
        with LoggerManager(labels={'key': 'out'}):
            self.assertEqual(LoggerManager.labels.get(), {'key': 'out'})
            with LoggerManager(labels={'key': 'in'}):
                self.assertEqual(LoggerManager.labels.get(), {'key': 'in'})
            self.assertEqual(LoggerManager.labels.get(), {'key': 'out'})


class LoggerTestCase(TestCase):

    def test_logger_return(self):
        log = Logger('everysk-log-stdout-test')
        self.assertIsInstance(log.log, PythonLogger)

    def test_log_level(self):
        log = Logger(name='everysk-log-stdout-test', level=INFO)
        self.assertEqual(log.log.level, INFO)
        log = Logger(name='everysk-log-stdout-test', level=DEBUG)
        self.assertEqual(log.log.level, DEBUG)

    def test_log_propagate(self):
        log = Logger('everysk-log-stdout-test')
        self.assertFalse(log.log.propagate)

    def test_stdout_handler(self):
        log = Logger('everysk-log-stdout-test')
        handler = log.log.handlers[0]
        self.assertEqual(type(handler), StreamHandler)

    @mock.patch.object(log_module, 'StructuredLogHandler')
    def test_gcp_handler(self, StructuredLogHandler: mock.MagicMock): # pylint: disable=invalid-name
        log = Logger('everysk-log-gcp-test', google_project='test', labels={'request_id': 'uuid'})
        StructuredLogHandler.assert_called_once_with(project_id='test', labels={'request_id': 'uuid'})
        handler = log.log.handlers[0]
        self.assertEqual(handler, StructuredLogHandler.return_value)

    def test_handler_level(self):
        log = Logger(name='everysk-log-stdout-test', level=INFO)
        handler = log.log.handlers[0]
        self.assertEqual(handler.level, INFO)
        log = Logger(name='everysk-log-stdout-test', level=DEBUG)
        handler = log.log.handlers[0]
        self.assertEqual(handler.level, DEBUG)

    def test_handler_format(self):
        log = Logger(name='everysk-log-stdout-test', level=INFO)
        handler = log.log.handlers[0]
        self.assertEqual(
            handler.formatter._fmt,
            '%(asctime)s - %(levelname)s - %(labels)s - %(message)s'
        )

    @mock.patch.object(log_module.logging, 'getLogger')
    def test_context_manager_logger(self, getLogger: mock.MagicMock): # pylint: disable=invalid-name
        log = Logger(name='everysk-log-context-test')
        log.debug('MSG')
        with LoggerManager(labels={'key': 'out'}):
            log.debug('MSG')

            with LoggerManager(labels={'key': 'in'}):
                log.debug('MSG')

            log.debug('MSG')

        log.debug('MSG')

        getLogger.return_value.debug.assert_has_calls([
            mock.call('MSG', extra={'labels': {}}, stacklevel=2),
            mock.call('MSG', extra={'labels': {'key': 'out'}}, stacklevel=2),
            mock.call('MSG', extra={'labels': {'key': 'in'}}, stacklevel=2),
            mock.call('MSG', extra={'labels': {'key': 'out'}}, stacklevel=2),
            mock.call('MSG', extra={'labels': {}}, stacklevel=2),
        ])

    def test_get_labels_default(self):
        log = Logger(name='everysk-log-context-test')
        self.assertDictEqual(log.get_labels(), {})

    def test_get_labels_init(self):
        log = Logger(name='everysk-log-context-test', labels={'key': 'out'})
        self.assertDictEqual(log.get_labels(), {'key': 'out'})

    def test_get_labels_context(self):
        log = Logger(name='everysk-log-context-test', labels={'key': 'out'})
        with LoggerManager(labels={'key': 'in'}):
            self.assertDictEqual(log.get_labels(), {'key': 'in'})

        self.assertDictEqual(log.get_labels(), {'key': 'out'})

    def test_set_value_error(self):
        log = Logger(name='everysk-log-set-value-test')
        with self.assertRaisesRegex(ValueError, r"The \"attr\" attribute must be a \<class 'str'\>."):
            log._set_value(name='attr', value=1, default='def', klass=str)

    def test_set_value_default(self):
        log = Logger(name='everysk-log-set-value-test')
        log._set_value(name='attr', value=Undefined, default='def', klass=str)
        self.assertEqual(log.attr, 'def') # pylint: disable=no-member

    @mock.patch.object(log_module.logging, 'getLogger')
    def test_critical(self, getLogger: mock.MagicMock): # pylint: disable=invalid-name
        log = Logger(name='everysk-log-msg-test')
        log.critical('msg %s', 1)
        getLogger.return_value.critical.assert_called_once_with('msg %s', 1, extra={'labels': {}}, stacklevel=2)

    @mock.patch.object(log_module.logging, 'getLogger')
    def test_debug(self, getLogger: mock.MagicMock): # pylint: disable=invalid-name
        log = Logger(name='everysk-log-msg-test')
        log.debug('msg %s', 1)
        getLogger.return_value.debug.assert_called_once_with('msg %s', 1, extra={'labels': {}}, stacklevel=2)

    @mock.patch.object(log_module.logging, 'getLogger')
    def test_error(self, getLogger: mock.MagicMock): # pylint: disable=invalid-name
        log = Logger(name='everysk-log-msg-test')
        log.error('msg %s', 1)
        getLogger.return_value.error.assert_called_once_with('msg %s', 1, extra={'labels': {}}, stacklevel=2)

    @mock.patch.object(log_module.logging, 'getLogger')
    def test_exception(self, getLogger: mock.MagicMock): # pylint: disable=invalid-name
        log = Logger(name='everysk-log-msg-test')
        log.exception('msg %s', 1)
        getLogger.return_value.exception.assert_called_once_with('msg %s', 1, extra={'labels': {}}, stacklevel=2)

    @mock.patch.object(log_module.logging, 'getLogger')
    def test_info(self, getLogger: mock.MagicMock): # pylint: disable=invalid-name
        log = Logger(name='everysk-log-msg-test')
        log.info('msg %s', 1)
        getLogger.return_value.info.assert_called_once_with('msg %s', 1, extra={'labels': {}}, stacklevel=2)

    @mock.patch.object(log_module.logging, 'getLogger')
    def test_warning(self, getLogger: mock.MagicMock): # pylint: disable=invalid-name
        log = Logger(name='everysk-log-msg-test')
        log.warning('msg %s', 1)
        getLogger.return_value.warning.assert_called_once_with('msg %s', 1, extra={'labels': {}}, stacklevel=2)

    def test_default_value_name(self):
        # https://everysk.atlassian.net/browse/COD-3212
        log = Logger()
        self.assertEqual(log.name, 'everysk-root-log')
        with self.assertRaisesRegex(ValueError, 'The name of the log could not be "root".'):
            log = Logger(name='root')

    def test_get_kwargs(self):
        log = Logger('everysk-log-kwargs')
        self.assertDictEqual(log._get_kwargs(), {'extra': {'labels': {}}, 'stacklevel': 2})
        log = Logger('everysk-log-kwargs', labels={'key': 'value'})
        self.assertDictEqual(log._get_kwargs(), {'extra': {'labels': {'key': 'value'}}, 'stacklevel': 2})


@mock.patch.object(log_module.logging, 'getLogger')
class LoggerDeprecatedTestCase(TestCase):
    # pylint: disable=invalid-name

    def test_send_message(self, getLogger: mock.MagicMock):
        log = Logger(name='everysk-log-deprecated')
        log.deprecated('Teste')
        getLogger.return_value.warning.assert_called_once_with('DeprecationWarning: Teste', extra={'labels': {}}, stacklevel=2)

    def test_send_message_once(self, getLogger: mock.MagicMock):
        log = Logger(name='everysk-log-deprecated')
        log.deprecated('Teste 1', show_once=True)
        log.deprecated('Teste 1', show_once=True)
        getLogger.return_value.warning.assert_called_once_with('DeprecationWarning: Teste 1', extra={'labels': {}}, stacklevel=2)

    def test_send_message_twice(self, getLogger: mock.MagicMock):
        log = Logger(name='everysk-log-deprecated')
        log.deprecated('Teste 2', show_once=False)
        log.deprecated('Teste 2', show_once=False)
        getLogger.return_value.warning.assert_has_calls([
            mock.call('DeprecationWarning: Teste 2', extra={'labels': {}}, stacklevel=2),
            mock.call('DeprecationWarning: Teste 2', extra={'labels': {}}, stacklevel=2),
        ])


@mock.patch.object(slack_module, 'Slack')
class LoggerSlackTestCase(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.log = Logger(name='everysk-lib-test-slack-message')

    def setUp(self) -> None:
        self.old_profile = settings.PROFILE
        self.old_slack_url = settings.SLACK_URL
        settings.PROFILE = 'PROD'
        settings.SLACK_URL = 'http://localhost'
        # This removes the module from the cache
        sys.modules.pop('unittest')

    def tearDown(self) -> None:
        settings.PROFILE = self.old_profile
        settings.SLACK_URL = self.old_slack_url
        # Put back the module
        import unittest # pylint: disable=unused-import, import-outside-toplevel

    def test_danger(self, slack: mock.MagicMock):
        with mock.patch.object(self.log.log, 'error') as error:
            self.log.slack(title='Error', message='Error message', color='danger')
            error.assert_called_once_with('Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=2)
        slack.assert_called_once_with(title='Error', message='Error message', color='danger', url=settings.SLACK_URL)

    def test_warning(self, slack: mock.MagicMock):
        with mock.patch.object(self.log.log, 'warning') as warning:
            self.log.slack(title='Warning', message='Warning message', color='warning')
            warning.assert_called_once_with('Slack message: Warning -> Warning message', extra={'labels': {}}, stacklevel=2)
        slack.assert_called_once_with(title='Warning', message='Warning message', color='warning', url=settings.SLACK_URL)

    def test_success(self, slack: mock.MagicMock):
        with mock.patch.object(self.log.log, 'info') as info:
            self.log.slack(title='Success', message='Success message', color='success')
            info.assert_called_once_with('Slack message: Success -> Success message', extra={'labels': {}}, stacklevel=2)
        slack.assert_called_once_with(title='Success', message='Success message', color='success', url=settings.SLACK_URL)

    def test_not_url(self, slack: mock.MagicMock):
        settings.SLACK_URL = None
        with mock.patch.object(self.log.log, 'error') as error:
            self.log.slack(title='Error', message='Error message', color='danger')
            error.assert_called_once_with('Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=2)
        slack.assert_not_called()

    def test_not_prod(self, slack: mock.MagicMock):
        settings.PROFILE = 'DEV'
        with mock.patch.object(self.log.log, 'error') as error:
            self.log.slack(title='Error', message='Error message', color='danger')
            error.assert_called_once_with('Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=2)
        slack.assert_not_called()

    def test_in_tests(self, slack: mock.MagicMock):
        import unittest # pylint: disable=unused-import, import-outside-toplevel
        with mock.patch.object(self.log.log, 'error') as error:
            self.log.slack(title='Error', message='Error message', color='danger')
            error.assert_called_once_with('Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=2)
        slack.assert_not_called()

    def test_init_url(self, slack: mock.MagicMock):
        with mock.patch.object(self.log.log, 'error') as error:
            log = Logger(name='everysk-lib-test-slack-message', slack_url='http://example.com')
            log.slack(title='Error', message='Error message', color='danger')
            error.assert_called_once_with('Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=2)
        slack.assert_called_once_with(title='Error', message='Error message', color='danger', url='http://example.com')

    def test_method_url(self, slack: mock.MagicMock):
        with mock.patch.object(self.log.log, 'error') as error:
            self.log.slack(title='Error', message='Error message', color='danger', url='http://google.com')
            error.assert_called_once_with('Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=2)
        slack.assert_called_once_with(title='Error', message='Error message', color='danger', url='http://google.com')
